@@ -1,29 +1,21 @@
 /*
-    Copyright 2005-2014 Intel Corporation.  All Rights Reserved.
+    Copyright (c) 2005-2017 Intel Corporation
 
-    This file is part of Threading Building Blocks.
+    Licensed under the Apache License, Version 2.0 (the "License");
+    you may not use this file except in compliance with the License.
+    You may obtain a copy of the License at
+
+        http://www.apache.org/licenses/LICENSE-2.0
+
+    Unless required by applicable law or agreed to in writing, software
+    distributed under the License is distributed on an "AS IS" BASIS,
+    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+    See the License for the specific language governing permissions and
+    limitations under the License.
 
-    Threading Building Blocks is free software; you can redistribute it
-    and/or modify it under the terms of the GNU General Public License
-    version 2 as published by the Free Software Foundation.
 
-    Threading Building Blocks is distributed in the hope that it will be
-    useful, but WITHOUT ANY WARRANTY; without even the implied warranty
-    of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-    GNU General Public License for more details.
 
-    You should have received a copy of the GNU General Public License
-    along with Threading Building Blocks; if not, write to the Free Software
-    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 
-    As a special exception, you may use this file as part of a free software
-    library without restriction.  Specifically, if other files instantiate
-    templates or use macros or inline functions from this file, or you compile
-    this file and link it with other files to produce an executable, this
-    file does not by itself cause the resulting executable to be covered by
-    the GNU General Public License.  This exception does not however
-    invalidate any other reasons why the executable file might be covered by
-    the GNU General Public License.
 */
 
 #include "tbb/tbb_stddef.h"
@@ -35,19 +27,8 @@
 #include "concurrent_monitor.h"
 #include "itt_notify.h"
 #include <new>
-
-#if !TBB_USE_EXCEPTIONS && _MSC_VER
-    // Suppress "C++ exception handler used, but unwind semantics are not enabled" warning in STL headers
-    #pragma warning (push)
-    #pragma warning (disable: 4530)
-#endif
-
 #include <cstring>   // for memset()
 
-#if !TBB_USE_EXCEPTIONS && _MSC_VER
-    #pragma warning (pop)
-#endif
-
 using namespace std;
 
 #if defined(_MSC_VER) && defined(_Wp64)
@@ -82,14 +63,18 @@ struct micro_queue {
 
     spin_mutex page_mutex;
 
-    void push( const void* item, ticket k, concurrent_queue_base& base );
+    void push( const void* item, ticket k, concurrent_queue_base& base,
+               concurrent_queue_base::copy_specifics op_type );
+
     void abort_push( ticket k, concurrent_queue_base& base );
 
     bool pop( void* dst, ticket k, concurrent_queue_base& base );
 
-    micro_queue& assign( const micro_queue& src, concurrent_queue_base& base );
+    micro_queue& assign( const micro_queue& src, concurrent_queue_base& base,
+                         concurrent_queue_base::copy_specifics op_type );
 
-    page* make_copy ( concurrent_queue_base& base, const page* src_page, size_t begin_in_page, size_t end_in_page, ticket& g_index ) ;
+    page* make_copy ( concurrent_queue_base& base, const page* src_page, size_t begin_in_page,
+                      size_t end_in_page, ticket& g_index, concurrent_queue_base::copy_specifics op_type ) ;
 
     void make_invalid( ticket k );
 };
@@ -162,6 +147,8 @@ public:
         return array[index(k)];
     }
 
+    atomic<unsigned> abort_counter;
+
     //! Value for effective_capacity that denotes unbounded queue.
     static const ptrdiff_t infinite_capacity = ptrdiff_t(~size_t(0)/2);
 };
@@ -172,12 +159,13 @@ public:
     #pragma warning( disable: 4146 )
 #endif
 
-static void* invalid_page;
+static void* static_invalid_page;
 
 //------------------------------------------------------------------------
 // micro_queue
 //------------------------------------------------------------------------
-void micro_queue::push( const void* item, ticket k, concurrent_queue_base& base ) {
+void micro_queue::push( const void* item, ticket k, concurrent_queue_base& base,
+                        concurrent_queue_base::copy_specifics op_type ) {
     k &= -concurrent_queue_rep::n_queue;
     page* p = NULL;
     // find index on page where we would put the data
@@ -188,6 +176,7 @@ void micro_queue::push( const void* item, ticket k, concurrent_queue_base& base
         } __TBB_CATCH(...) {
             ++base.my_rep->n_invalid_entries;
             make_invalid( k );
+            __TBB_RETHROW();
         }
         p->mask = 0;
         p->next = NULL;
@@ -218,10 +207,15 @@ void micro_queue::push( const void* item, ticket k, concurrent_queue_base& base
         p = tail_page;
         ITT_NOTIFY( sync_acquired, p );
         __TBB_TRY {
-            base.copy_item( *p, index, item );
+            if( concurrent_queue_base::copy == op_type ) {
+                base.copy_item( *p, index, item );
+            } else {
+                __TBB_ASSERT( concurrent_queue_base::move == op_type, NULL );
+                static_cast<concurrent_queue_base_v8&>(base).move_item( *p, index, item );
+            }
         }  __TBB_CATCH(...) {
             ++base.my_rep->n_invalid_entries;
-            tail_counter += concurrent_queue_rep::n_queue; 
+            tail_counter += concurrent_queue_rep::n_queue;
             __TBB_RETHROW();
         }
         ITT_NOTIFY( sync_releasing, p );
@@ -231,29 +225,29 @@ void micro_queue::push( const void* item, ticket k, concurrent_queue_base& base
     else // no item; this was called from abort_push
         ++base.my_rep->n_invalid_entries;
 
-    tail_counter += concurrent_queue_rep::n_queue; 
+    tail_counter += concurrent_queue_rep::n_queue;
 }
 
 
 void micro_queue::abort_push( ticket k, concurrent_queue_base& base ) {
-    push(NULL, k, base);
-} 
+    push(NULL, k, base, concurrent_queue_base::copy);
+}
 
 bool micro_queue::pop( void* dst, ticket k, concurrent_queue_base& base ) {
     k &= -concurrent_queue_rep::n_queue;
     spin_wait_until_eq( head_counter, k );
     spin_wait_while_eq( tail_counter, k );
-    page& p = *head_page;
-    __TBB_ASSERT( &p, NULL );
+    page *p = head_page;
+    __TBB_ASSERT( p, NULL );
     size_t index = modulo_power_of_two( k/concurrent_queue_rep::n_queue, base.items_per_page );
     bool success = false;
     {
-        micro_queue_pop_finalizer finalizer( *this, base, k+concurrent_queue_rep::n_queue, index==base.items_per_page-1 ? &p : NULL );
-        if( p.mask & uintptr_t(1)<<index ) {
+        micro_queue_pop_finalizer finalizer( *this, base, k+concurrent_queue_rep::n_queue, index==base.items_per_page-1 ? p : NULL );
+        if( p->mask & uintptr_t(1)<<index ) {
             success = true;
             ITT_NOTIFY( sync_acquired, dst );
             ITT_NOTIFY( sync_acquired, head_page );
-            base.assign_and_destroy_item( dst, p, index );
+            base.assign_and_destroy_item( dst, *p, index );
             ITT_NOTIFY( sync_releasing, head_page );
         } else {
             --base.my_rep->n_invalid_entries;
@@ -262,11 +256,11 @@ bool micro_queue::pop( void* dst, ticket k, concurrent_queue_base& base ) {
     return success;
 }
 
-micro_queue& micro_queue::assign( const micro_queue& src, concurrent_queue_base& base )
+micro_queue& micro_queue::assign( const micro_queue& src, concurrent_queue_base& base,
+                                  concurrent_queue_base::copy_specifics op_type )
 {
     head_counter = src.head_counter;
     tail_counter = src.tail_counter;
-    page_mutex   = src.page_mutex;
 
     const page* srcp = src.head_page;
     if( srcp ) {
@@ -276,12 +270,12 @@ micro_queue& micro_queue::assign( const micro_queue& src, concurrent_queue_base&
             size_t index = modulo_power_of_two( head_counter/concurrent_queue_rep::n_queue, base.items_per_page );
             size_t end_in_first_page = (index+n_items<base.items_per_page)?(index+n_items):base.items_per_page;
 
-            head_page = make_copy( base, srcp, index, end_in_first_page, g_index );
+            head_page = make_copy( base, srcp, index, end_in_first_page, g_index, op_type );
             page* cur_page = head_page;
 
             if( srcp != src.tail_page ) {
                 for( srcp = srcp->next; srcp!=src.tail_page; srcp=srcp->next ) {
-                    cur_page->next = make_copy( base, srcp, 0, base.items_per_page, g_index );
+                    cur_page->next = make_copy( base, srcp, 0, base.items_per_page, g_index, op_type );
                     cur_page = cur_page->next;
                 }
 
@@ -290,12 +284,13 @@ micro_queue& micro_queue::assign( const micro_queue& src, concurrent_queue_base&
                 size_t last_index = modulo_power_of_two( tail_counter/concurrent_queue_rep::n_queue, base.items_per_page );
                 if( last_index==0 ) last_index = base.items_per_page;
 
-                cur_page->next = make_copy( base, srcp, 0, last_index, g_index );
+                cur_page->next = make_copy( base, srcp, 0, last_index, g_index, op_type );
                 cur_page = cur_page->next;
             }
             tail_page = cur_page;
         } __TBB_CATCH(...) {
             make_invalid( g_index );
+            __TBB_RETHROW();
         }
     } else {
         head_page = tail_page = NULL;
@@ -303,14 +298,22 @@ micro_queue& micro_queue::assign( const micro_queue& src, concurrent_queue_base&
     return *this;
 }
 
-concurrent_queue_base::page* micro_queue::make_copy( concurrent_queue_base& base, const concurrent_queue_base::page* src_page, size_t begin_in_page, size_t end_in_page, ticket& g_index )
+concurrent_queue_base::page* micro_queue::make_copy( concurrent_queue_base& base,
+    const concurrent_queue_base::page* src_page, size_t begin_in_page, size_t end_in_page,
+    ticket& g_index, concurrent_queue_base::copy_specifics op_type )
 {
     page* new_page = base.allocate_page();
     new_page->next = NULL;
     new_page->mask = src_page->mask;
     for( ; begin_in_page!=end_in_page; ++begin_in_page, ++g_index )
-        if( new_page->mask & uintptr_t(1)<<begin_in_page )
-            base.copy_page_item( *new_page, begin_in_page, *src_page, begin_in_page );
+        if( new_page->mask & uintptr_t(1)<<begin_in_page ) {
+            if( concurrent_queue_base::copy == op_type ) {
+                base.copy_page_item( *new_page, begin_in_page, *src_page, begin_in_page );
+            } else {
+                __TBB_ASSERT( concurrent_queue_base::move == op_type, NULL );
+                static_cast<concurrent_queue_base_v8&>(base).move_page_item( *new_page, begin_in_page, *src_page, begin_in_page );
+            }
+        }
     return new_page;
 }
 
@@ -318,17 +321,16 @@ void micro_queue::make_invalid( ticket k )
 {
     static concurrent_queue_base::page dummy = {static_cast<page*>((void*)1), 0};
     // mark it so that no more pushes are allowed.
-    invalid_page = &dummy;
+    static_invalid_page = &dummy;
     {
         spin_mutex::scoped_lock lock( page_mutex );
         tail_counter = k+concurrent_queue_rep::n_queue+1;
         if( page* q = tail_page )
-            q->next = static_cast<page*>(invalid_page);
+            q->next = static_cast<page*>(static_invalid_page);
         else
-            head_page = static_cast<page*>(invalid_page);
-        tail_page = static_cast<page*>(invalid_page);
+            head_page = static_cast<page*>(static_invalid_page);
+        tail_page = static_cast<page*>(static_invalid_page);
     }
-    __TBB_RETHROW();
 }
 
 #if _MSC_VER && !defined(__INTEL_COMPILER)
@@ -347,10 +349,10 @@ concurrent_queue_base_v3::concurrent_queue_base_v3( size_t item_sz ) {
                      1;
     my_capacity = size_t(-1)/(item_sz>1 ? item_sz : 2);
     my_rep = cache_aligned_allocator<concurrent_queue_rep>().allocate(1);
-    __TBB_ASSERT( (size_t)my_rep % NFS_GetLineSize()==0, "alignment error" );
-    __TBB_ASSERT( (size_t)&my_rep->head_counter % NFS_GetLineSize()==0, "alignment error" );
-    __TBB_ASSERT( (size_t)&my_rep->tail_counter % NFS_GetLineSize()==0, "alignment error" );
-    __TBB_ASSERT( (size_t)&my_rep->array % NFS_GetLineSize()==0, "alignment error" );
+    __TBB_ASSERT( is_aligned(my_rep, NFS_GetLineSize()), "alignment error" );
+    __TBB_ASSERT( is_aligned(&my_rep->head_counter, NFS_GetLineSize()), "alignment error" );
+    __TBB_ASSERT( is_aligned(&my_rep->tail_counter, NFS_GetLineSize()), "alignment error" );
+    __TBB_ASSERT( is_aligned(&my_rep->array, NFS_GetLineSize()), "alignment error" );
     memset(my_rep,0,sizeof(concurrent_queue_rep));
     new ( &my_rep->items_avail ) concurrent_monitor();
     new ( &my_rep->slots_avail ) concurrent_monitor();
@@ -365,7 +367,16 @@ concurrent_queue_base_v3::~concurrent_queue_base_v3() {
 }
 
 void concurrent_queue_base_v3::internal_push( const void* src ) {
+    internal_insert_item( src, copy );
+}
+
+void concurrent_queue_base_v8::internal_push_move( const void* src ) {
+   internal_insert_item( src, move );
+}
+
+void concurrent_queue_base_v3::internal_insert_item( const void* src, copy_specifics op_type ) {
     concurrent_queue_rep& r = *my_rep;
+    unsigned old_abort_counter = r.abort_counter;
     ticket k = r.tail_counter++;
     ptrdiff_t e = my_capacity;
 #if DO_ITT_NOTIFY
@@ -383,6 +394,10 @@ void concurrent_queue_base_v3::internal_push( const void* src ) {
         r.slots_avail.prepare_wait( thr_ctx, ((ptrdiff_t)(k-e)) );
         while( (ptrdiff_t)(k-r.head_counter)>=const_cast<volatile ptrdiff_t&>(e = my_capacity) ) {
             __TBB_TRY {
+                if( r.abort_counter!=old_abort_counter ) {
+                    r.slots_avail.cancel_wait( thr_ctx );
+                    throw_exception( eid_user_abort );
+                }
                 slept = r.slots_avail.commit_wait( thr_ctx );
             } __TBB_CATCH( tbb::user_abort& ) {
                 r.choose(k).abort_push(k, *this);
@@ -398,7 +413,7 @@ void concurrent_queue_base_v3::internal_push( const void* src ) {
     }
     ITT_NOTIFY( sync_acquired, &sync_prepare_done );
     __TBB_ASSERT( (ptrdiff_t)(k-r.head_counter)<my_capacity, NULL);
-    r.choose( k ).push( src, k, *this );
+    r.choose( k ).push( src, k, *this, op_type );
     r.items_avail.notify( predicate_leq(k) );
 }
 
@@ -408,6 +423,8 @@ void concurrent_queue_base_v3::internal_pop( void* dst ) {
 #if DO_ITT_NOTIFY
     bool sync_prepare_done = false;
 #endif
+    unsigned old_abort_counter = r.abort_counter;
+    // This loop is a single pop operation; abort_counter should not be re-read inside
     do {
         k=r.head_counter++;
         if ( (ptrdiff_t)(r.tail_counter-k)<=0 ) { // queue is empty
@@ -422,6 +439,10 @@ void concurrent_queue_base_v3::internal_pop( void* dst ) {
             r.items_avail.prepare_wait( thr_ctx, k );
             while( (ptrdiff_t)(r.tail_counter-k)<=0 ) {
                 __TBB_TRY {
+                    if( r.abort_counter!=old_abort_counter ) {
+                        r.items_avail.cancel_wait( thr_ctx );
+                        throw_exception( eid_user_abort );
+                    }
                     slept = r.items_avail.commit_wait( thr_ctx );
                 } __TBB_CATCH( tbb::user_abort& ) {
                     r.head_counter--;
@@ -444,6 +465,7 @@ void concurrent_queue_base_v3::internal_pop( void* dst ) {
 
 void concurrent_queue_base_v3::internal_abort() {
     concurrent_queue_rep& r = *my_rep;
+    ++r.abort_counter;
     r.items_avail.abort_all();
     r.slots_avail.abort_all();
 }
@@ -473,6 +495,14 @@ bool concurrent_queue_base_v3::internal_pop_if_present( void* dst ) {
 }
 
 bool concurrent_queue_base_v3::internal_push_if_not_full( const void* src ) {
+    return internal_insert_if_not_full( src, copy );
+}
+
+bool concurrent_queue_base_v8::internal_push_move_if_not_full( const void* src ) {
+    return internal_insert_if_not_full( src, move );
+}
+
+bool concurrent_queue_base_v3::internal_insert_if_not_full( const void* src, copy_specifics op_type ) {
     concurrent_queue_rep& r = *my_rep;
     ticket k = r.tail_counter;
     for(;;) {
@@ -487,8 +517,7 @@ bool concurrent_queue_base_v3::internal_push_if_not_full( const void* src ) {
             break;
         // Another thread claimed the slot, so retry.
     }
-    r.choose(k).push(src,k,*this);
-
+    r.choose(k).push(src, k, *this, op_type);
     r.items_avail.notify( predicate_leq(k) );
     return true;
 }
@@ -515,7 +544,7 @@ void concurrent_queue_base_v3::internal_finish_clear() {
         page* tp = my_rep->array[i].tail_page;
         __TBB_ASSERT( my_rep->array[i].head_page==tp, "at most one page should remain" );
         if( tp!=NULL) {
-            if( tp!=invalid_page ) deallocate_page( tp );
+            if( tp!=static_invalid_page ) deallocate_page( tp );
             my_rep->array[i].tail_page = NULL;
         }
     }
@@ -525,7 +554,7 @@ void concurrent_queue_base_v3::internal_throw_exception() const {
     throw_exception( eid_bad_alloc );
 }
 
-void concurrent_queue_base_v3::assign( const concurrent_queue_base& src ) {
+void concurrent_queue_base_v3::internal_assign( const concurrent_queue_base& src, copy_specifics op_type ) {
     items_per_page = src.items_per_page;
     my_capacity = src.my_capacity;
 
@@ -533,15 +562,24 @@ void concurrent_queue_base_v3::assign( const concurrent_queue_base& src ) {
     my_rep->head_counter = src.my_rep->head_counter;
     my_rep->tail_counter = src.my_rep->tail_counter;
     my_rep->n_invalid_entries = src.my_rep->n_invalid_entries;
+    my_rep->abort_counter = src.my_rep->abort_counter;
 
     // copy micro_queues
     for( size_t i = 0; i<my_rep->n_queue; ++i )
-        my_rep->array[i].assign( src.my_rep->array[i], *this);
+        my_rep->array[i].assign( src.my_rep->array[i], *this, op_type );
 
     __TBB_ASSERT( my_rep->head_counter==src.my_rep->head_counter && my_rep->tail_counter==src.my_rep->tail_counter,
             "the source concurrent queue should not be concurrently modified." );
 }
 
+void concurrent_queue_base_v3::assign( const concurrent_queue_base& src ) {
+    internal_assign( src, copy );
+}
+
+void concurrent_queue_base_v8::move_content( concurrent_queue_base_v8& src ) {
+    internal_assign( src, move );
+}
+
 //------------------------------------------------------------------------
 // concurrent_queue_iterator_rep
 //------------------------------------------------------------------------
