@@ -1,21 +1,21 @@
 /*
-    Copyright 2005-2014 Intel Corporation.  All Rights Reserved.
+    Copyright (c) 2005-2017 Intel Corporation
+
+    Licensed under the Apache License, Version 2.0 (the "License");
+    you may not use this file except in compliance with the License.
+    You may obtain a copy of the License at
+
+        http://www.apache.org/licenses/LICENSE-2.0
+
+    Unless required by applicable law or agreed to in writing, software
+    distributed under the License is distributed on an "AS IS" BASIS,
+    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+    See the License for the specific language governing permissions and
+    limitations under the License.
+
+
 
-    This file is part of Threading Building Blocks. Threading Building Blocks is free software;
-    you can redistribute it and/or modify it under the terms of the GNU General Public License
-    version 2  as  published  by  the  Free Software Foundation.  Threading Building Blocks is
-    distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
-    implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
-    See  the GNU General Public License for more details.   You should have received a copy of
-    the  GNU General Public License along with Threading Building Blocks; if not, write to the
-    Free Software Foundation, Inc.,  51 Franklin St,  Fifth Floor,  Boston,  MA 02110-1301 USA
 
-    As a special exception,  you may use this file  as part of a free software library without
-    restriction.  Specifically,  if other files instantiate templates  or use macros or inline
-    functions from this file, or you compile this file and link it with other files to produce
-    an executable,  this file does not by itself cause the resulting executable to be covered
-    by the GNU General Public License. This exception does not however invalidate any other
-    reasons why the executable file might be covered by the GNU General Public License.
 */
 
 #ifndef _TBB_scheduler_H
@@ -28,6 +28,8 @@
 #include "itt_notify.h"
 #include "../rml/include/rml_tbb.h"
 
+#include "intrusive_list.h"
+
 #if __TBB_SURVIVE_THREAD_SWITCH
 #include "cilk-tbb-interop.h"
 #endif /* __TBB_SURVIVE_THREAD_SWITCH */
@@ -36,23 +38,26 @@ namespace tbb {
 namespace internal {
 
 template<typename SchedulerTraits> class custom_scheduler;
-struct nested_arena_context;
 
 //------------------------------------------------------------------------
 // generic_scheduler
 //------------------------------------------------------------------------
 
-#if __TBB_TASK_GROUP_CONTEXT
-struct scheduler_list_node_t {
-    scheduler_list_node_t *my_prev,
-                          *my_next;
-};
-#endif /* __TBB_TASK_GROUP_CONTEXT */
-
 #define EmptyTaskPool ((task**)0)
 #define LockedTaskPool ((task**)~(intptr_t)0)
 
-#define LockedMaster ((generic_scheduler*)~(intptr_t)0)
+//! Bit-field representing properties of a sheduler
+struct scheduler_properties {
+    static const bool worker = false;
+    static const bool master = true;
+    //! Indicates that a scheduler acts as a master or a worker.
+    bool type : 1;
+    //! Indicates that a scheduler is on outermost level.
+    /**  Note that the explicit execute method will set this property. **/
+    bool outermost : 1;
+    //! Reserved bits
+    unsigned char : 6;
+};
 
 struct scheduler_state {
     //! Index of the arena slot the scheduler occupies now, or occupied last time.
@@ -64,14 +69,9 @@ struct scheduler_state {
     //! The arena that I own (if master) or am servicing at the moment (if worker)
     arena* my_arena;
 
-    //! Innermost task whose task::execute() is running.
+    //! Innermost task whose task::execute() is running. A dummy task on the outermost level.
     task* my_innermost_running_task;
 
-    //! Task, in the context of which the current TBB dispatch loop is running.
-    /** Outside of or in the outermost dispatch loop (not in a nested call to
-        wait_for_all) it is my_dummy_task for master threads, and NULL for workers. **/
-    task* my_dispatching_task;
-
     mail_inbox my_inbox;
 
     //! The mailbox id assigned to this scheduler.
@@ -79,16 +79,21 @@ struct scheduler_state {
         TODO: how are id's being garbage collected?
         TODO: master thread may enter arena and leave and then reenter.
                 We want to give it the same affinity_id upon reentry, if practical.
+        TODO: investigate if it makes sense to merge this field into scheduler_properties.
       */
     affinity_id my_affinity_id;
 
+    scheduler_properties my_properties;
+
 #if __TBB_SCHEDULER_OBSERVER
     //! Last observer in the global observers list processed by this scheduler
     observer_proxy* my_last_global_observer;
+#endif
 
+#if __TBB_ARENA_OBSERVER
     //! Last observer in the local observers list processed by this scheduler
     observer_proxy* my_last_local_observer;
-#endif /* __TBB_SCHEDULER_OBSERVER */
+#endif
 #if __TBB_TASK_PRIORITY
     //! Latest known highest priority of tasks in the market or arena.
     /** Master threads currently tracks only tasks in their arenas, while workers
@@ -107,7 +112,10 @@ struct scheduler_state {
     Class generic_scheduler is an abstract base class that contains most of the scheduler,
     except for tweaks specific to processors and tools (e.g. VTune).
     The derived template class custom_scheduler<SchedulerTraits> fills in the tweaks. */
-class generic_scheduler: public scheduler, public ::rml::job, public scheduler_state {
+class generic_scheduler: public scheduler
+                       , public ::rml::job
+                       , public intrusive_list_node
+                       , public scheduler_state {
 public: // almost every class in TBB uses generic_scheduler
 
     //! If sizeof(task) is <=quick_task_size, it is handled on a free list instead of malloc'd.
@@ -126,8 +134,7 @@ public: // almost every class in TBB uses generic_scheduler
 
     static const size_t null_arena_index = ~size_t(0);
 
-    // TODO: Rename into is_task_pool_published()
-    inline bool in_arena () const;
+    inline bool is_task_pool_published () const;
 
     inline bool is_local_task_pool_quiescent () const;
 
@@ -182,21 +189,17 @@ public: // almost every class in TBB uses generic_scheduler
 #endif
     }
 
-    //! Actions common to enter_arena and try_enter_arena
-    void do_enter_arena();
-
-    //! Used by workers to enter the arena 
+    //! Used by workers to enter the task pool
     /** Does not lock the task pool in case if arena slot has been successfully grabbed. **/
-    void enter_arena();
+    void publish_task_pool();
 
-    //! Leave the arena
-    /** Leaving arena automatically releases the task pool if it is locked. **/
-    void leave_arena();
+    //! Leave the task pool
+    /** Leaving task pool automatically releases the task pool if it is locked. **/
+    void leave_task_pool();
 
-    //! Resets head and tail indices to 0, and leaves arena
-    /** Argument specifies whether the task pool is currently locked by the owner
-        (via acquire_task_pool).**/
-    inline void reset_deque_and_leave_arena ( bool locked );
+    //! Resets head and tail indices to 0, and leaves task pool
+    /** The task pool must be locked by the owner (via acquire_task_pool).**/
+    inline void reset_task_pool_and_leave ();
 
     //! Locks victim's task pool, and returns pointer to it. The pointer can be NULL.
     /** Garbles victim_arena_slot->task_pool for the duration of the lock. **/
@@ -229,18 +232,29 @@ public: // almost every class in TBB uses generic_scheduler
 
     //! Get a task from the local pool.
     /** Called only by the pool owner.
-        Returns the pointer to the task or NULL if the pool is empty. 
-        In the latter case compacts the pool. **/
-    task* get_task();
+        Returns the pointer to the task or NULL if a suitable task is not found.
+        Resets the pool if it is empty. **/
+    task* get_task( __TBB_ISOLATION_EXPR( isolation_tag isolation ) );
 
+    //! Get a task from the local pool at specified location T.
+    /** Returns the pointer to the task or NULL if the task cannot be executed,
+        e.g. proxy has been deallocated or isolation constraint is not met.
+        tasks_omitted tells if some tasks have been omitted.
+        Called only by the pool owner. The caller should guarantee that the
+        position T is not available for a thief. **/
+#if __TBB_TASK_ISOLATION
+    task* get_task( size_t T, isolation_tag isolation, bool& tasks_omitted );
+#else
+    task* get_task( size_t T );
+#endif /* __TBB_TASK_ISOLATION */
     //! Attempt to get a task from the mailbox.
-    /** Gets a task only if it has not been executed by its sender or a thief 
+    /** Gets a task only if it has not been executed by its sender or a thief
         that has stolen it from the sender's task pool. Otherwise returns NULL.
 
-        This method is intended to be used only by the thread extracting the proxy 
+        This method is intended to be used only by the thread extracting the proxy
         from its mailbox. (In contrast to local task pool, mailbox can be read only
         by its owner). **/
-    task* get_mailbox_task();
+    task* get_mailbox_task( __TBB_ISOLATION_EXPR( isolation_tag isolation ) );
 
     //! True if t is a task_proxy
     static bool is_proxy( const task& t ) {
@@ -248,7 +262,7 @@ public: // almost every class in TBB uses generic_scheduler
     }
 
     //! Steal task from another scheduler's ready pool.
-    task* steal_task( arena_slot& victim_arena_slot );
+    task* steal_task( __TBB_ISOLATION_ARG( arena_slot& victim_arena_slot, isolation_tag isolation ) );
 
     /** Initial size of the task deque sufficient to serve without reallocation
         4 nested parallel_for calls with iteration space of 65535 grains each. **/
@@ -260,10 +274,10 @@ public: // almost every class in TBB uses generic_scheduler
     size_t prepare_task_pool( size_t n );
 
     //! Initialize a scheduler for a master thread.
-    static generic_scheduler* create_master( arena& a );
+    static generic_scheduler* create_master( arena* a );
 
     //! Perform necessary cleanup when a master thread stops using TBB.
-    void cleanup_master();
+    bool cleanup_master( bool blocking_terminate );
 
     //! Initialize a scheduler for a worker thread.
     static generic_scheduler* create_worker( market& m, size_t index );
@@ -273,34 +287,30 @@ public: // almost every class in TBB uses generic_scheduler
 
 protected:
     template<typename SchedulerTraits> friend class custom_scheduler;
-    generic_scheduler( arena*, size_t index );
+    generic_scheduler( market & );
 
 public:
 #if TBB_USE_ASSERT > 1
     //! Check that internal data structures are in consistent state.
     /** Raises __TBB_ASSERT failure if inconsistency is found. */
-    void assert_task_pool_valid () const;
+    void assert_task_pool_valid() const;
 #else
     void assert_task_pool_valid() const {}
 #endif /* TBB_USE_ASSERT <= 1 */
 
-#if __TBB_TASK_ARENA
-    void nested_arena_entry(arena*, nested_arena_context &, bool);
-    void nested_arena_exit(nested_arena_context &);
+    void attach_arena( arena*, size_t index, bool is_master );
+    void nested_arena_entry( arena*, size_t );
+    void nested_arena_exit();
     void wait_until_empty();
-#endif
 
-    /*override*/ 
-    void spawn( task& first, task*& next );
+    void spawn( task& first, task*& next ) __TBB_override;
 
-    /*override*/ 
-    void spawn_root_and_wait( task& first, task*& next );
+    void spawn_root_and_wait( task& first, task*& next ) __TBB_override;
 
-    /*override*/ 
-    void enqueue( task&, void* reserved );
+    void enqueue( task&, void* reserved ) __TBB_override;
 
-    void local_spawn( task& first, task*& next );
-    void local_spawn_root_and_wait( task& first, task*& next );
+    void local_spawn( task* first, task*& next );
+    void local_spawn_root_and_wait( task* first, task*& next );
     virtual void local_wait_for_all( task& parent, task* child ) = 0;
 
     //! Destroy and deallocate this scheduler object
@@ -308,7 +318,7 @@ public:
 
     //! Allocate task object, either from the heap or a free list.
     /** Returns uninitialized task object with initialized prefix. */
-    task& allocate_task( size_t number_of_bytes, 
+    task& allocate_task( size_t number_of_bytes,
                        __TBB_CONTEXT_ARG(task* parent, task_group_context* context) );
 
     //! Put task on free list.
@@ -320,29 +330,22 @@ public:
     inline void deallocate_task( task& t );
 
     //! True if running on a worker thread, false otherwise.
-    inline bool is_worker();
+    inline bool is_worker() const;
+
+    //! True if the scheduler is on the outermost dispatch level.
+    inline bool outermost_level() const;
 
     //! True if the scheduler is on the outermost dispatch level in a master thread.
     /** Returns true when this scheduler instance is associated with an application
-        thread, and is not executing any TBB task. This includes being in a TBB 
+        thread, and is not executing any TBB task. This includes being in a TBB
         dispatch loop (one of wait_for_all methods) invoked directly from that thread. **/
     inline bool master_outermost_level () const;
 
     //! True if the scheduler is on the outermost dispatch level in a worker thread.
     inline bool worker_outermost_level () const;
 
-#if __TBB_TASK_GROUP_CONTEXT
-    //! Returns task group context used by this scheduler instance.
-    /** This context is associated with root tasks created by a master thread 
-        without explicitly specified context object outside of any running task.
-
-        Note that the default context of a worker thread is never accessed by
-        user code (directly or indirectly). **/
-    inline task_group_context* default_context ();
-#endif /* __TBB_TASK_GROUP_CONTEXT */
-
-    //! Returns number of worker threads in the arena this thread belongs to.
-    unsigned number_of_workers_in_my_arena();
+    //! Returns the concurrency limit of the current arena.
+    unsigned max_threads_in_arena();
 
 #if __TBB_COUNT_TASK_NODES
     intptr_t get_task_node_count( bool count_arena_workers = false );
@@ -351,21 +354,29 @@ public:
     //! Special value used to mark my_return_list as not taking any more entries.
     static task* plugged_return_list() {return (task*)(intptr_t)(-1);}
 
-    //! Number of small tasks that have been allocated by this scheduler. 
-    intptr_t my_small_task_count;
+    //! Number of small tasks that have been allocated by this scheduler.
+    __TBB_atomic intptr_t my_small_task_count;
 
     //! List of small tasks that have been returned to this scheduler by other schedulers.
+    // TODO IDEA: see if putting my_return_list on separate cache line improves performance
     task* my_return_list;
 
     //! Try getting a task from other threads (via mailbox, stealing, FIFO queue, orphans adoption).
     /** Returns obtained task or NULL if all attempts fail. */
-    virtual task* receive_or_steal_task( __TBB_atomic reference_count& completion_ref_count,
-                                         bool return_if_no_work ) = 0;
+    virtual task* receive_or_steal_task( __TBB_ISOLATION_ARG( __TBB_atomic reference_count& completion_ref_count, isolation_tag isolation ) ) = 0;
 
-    //! Free a small task t that that was allocated by a different scheduler 
-    void free_nonlocal_small_task( task& t ); 
+    //! Free a small task t that that was allocated by a different scheduler
+    void free_nonlocal_small_task( task& t );
 
 #if __TBB_TASK_GROUP_CONTEXT
+    //! Returns task group context used by this scheduler instance.
+    /** This context is associated with root tasks created by a master thread
+        without explicitly specified context object outside of any running task.
+
+        Note that the default context of a worker thread is never accessed by
+        user code (directly or indirectly). **/
+    inline task_group_context* default_context ();
+
     //! Padding isolating thread-local members from members that can be written to by other threads.
     char _padding1[NFS_MaxLineSize - sizeof(context_list_node_t)];
 
@@ -376,18 +387,18 @@ public:
     // TODO: check whether it can be deadly preempted and replace by spinning/sleeping mutex
     spin_mutex my_context_list_mutex;
 
-    //! Last state propagation epoch known to this thread 
+    //! Last state propagation epoch known to this thread
     /** Together with the_context_state_propagation_epoch constitute synchronization protocol
-        that keeps hot path of task group context construction destruction mostly 
+        that keeps hot path of task group context construction destruction mostly
         lock-free.
         When local epoch equals the global one, the state of task group contexts
         registered with this thread is consistent with that of the task group trees
         they belong to. **/
     uintptr_t my_context_state_propagation_epoch;
 
-    //! Flag indicating that a context is being destructed by its owner thread 
+    //! Flag indicating that a context is being destructed by its owner thread
     /** Together with my_nonlocal_ctx_list_update constitute synchronization protocol
-        that keeps hot path of context destruction (by the owner thread) mostly 
+        that keeps hot path of context destruction (by the owner thread) mostly
         lock-free. **/
     tbb::atomic<uintptr_t> my_local_ctx_list_update;
 
@@ -410,13 +421,17 @@ public:
 
     //! Searches offload area for top priority tasks and reloads found ones into primary task pool.
     /** Returns one of the found tasks or NULL. **/
-    task* reload_tasks ();
+    task* reload_tasks( __TBB_ISOLATION_EXPR( isolation_tag isolation ) );
 
-    task* reload_tasks ( task*& offloaded_tasks, task**& offloaded_task_list_link, intptr_t top_priority );
+    task* reload_tasks( task*& offloaded_tasks, task**& offloaded_task_list_link, __TBB_ISOLATION_ARG( intptr_t top_priority, isolation_tag isolation ) );
 
     //! Moves tasks with priority below the top one from primary task pool into offload area.
     /** Returns the next execution candidate task or NULL. **/
-    task* winnow_task_pool ();
+    task* winnow_task_pool ( __TBB_ISOLATION_EXPR( isolation_tag isolation ) );
+
+    //! Get a task from locked or empty pool in range [H0, T0). Releases or unlocks the task pool.
+    /** Returns the found task or NULL. **/
+    task *get_task_and_activate_task_pool( size_t H0 , __TBB_ISOLATION_ARG( size_t T0, isolation_tag isolation ) );
 
     //! Unconditionally moves the task into offload area.
     inline void offload_task ( task& t, intptr_t task_priority );
@@ -507,7 +522,7 @@ public:
 namespace tbb {
 namespace internal {
 
-inline bool generic_scheduler::in_arena () const {
+inline bool generic_scheduler::is_task_pool_published () const {
     __TBB_ASSERT(my_arena_slot, 0);
     return my_arena_slot->task_pool != EmptyTaskPool;
 }
@@ -528,12 +543,16 @@ inline bool generic_scheduler::is_quiescent_local_task_pool_reset () const {
     return __TBB_load_relaxed(my_arena_slot->head) == 0 && __TBB_load_relaxed(my_arena_slot->tail) == 0;
 }
 
+inline bool generic_scheduler::outermost_level () const {
+    return my_properties.outermost;
+}
+
 inline bool generic_scheduler::master_outermost_level () const {
-    return my_dispatching_task == my_dummy_task;
+    return !is_worker() && outermost_level();
 }
 
 inline bool generic_scheduler::worker_outermost_level () const {
-    return !my_dispatching_task;
+    return is_worker() && outermost_level();
 }
 
 #if __TBB_TASK_GROUP_CONTEXT
@@ -548,12 +567,13 @@ inline void generic_scheduler::attach_mailbox( affinity_id id ) {
     my_affinity_id = id;
 }
 
-inline bool generic_scheduler::is_worker() {
-    return my_arena_index != 0; //TODO: rework for multiple master
+inline bool generic_scheduler::is_worker() const {
+    return my_properties.type == scheduler_properties::worker;
 }
 
-inline unsigned generic_scheduler::number_of_workers_in_my_arena() {
-    return my_arena->my_max_num_workers;
+inline unsigned generic_scheduler::max_threads_in_arena() {
+    __TBB_ASSERT(my_arena, NULL);
+    return my_arena->my_num_slots;
 }
 
 //! Return task object to the memory allocator.
@@ -561,7 +581,7 @@ inline void generic_scheduler::deallocate_task( task& t ) {
 #if TBB_USE_ASSERT
     task_prefix& p = t.prefix();
     p.state = 0xFF;
-    p.extra_state = 0xFF; 
+    p.extra_state = 0xFF;
     poison_pointer(p.next);
 #endif /* TBB_USE_ASSERT */
     NFS_Free((char*)&t-task_prefix_reservation_size);
@@ -576,12 +596,11 @@ inline intptr_t generic_scheduler::get_task_node_count( bool count_arena_workers
 }
 #endif /* __TBB_COUNT_TASK_NODES */
 
-inline void generic_scheduler::reset_deque_and_leave_arena ( bool locked ) {
-    if ( !locked )
-        acquire_task_pool();
+inline void generic_scheduler::reset_task_pool_and_leave () {
+    __TBB_ASSERT( my_arena_slot->task_pool == LockedTaskPool, "Task pool must be locked when resetting task pool" );
     __TBB_store_relaxed( my_arena_slot->tail, 0 );
     __TBB_store_relaxed( my_arena_slot->head, 0 );
-    leave_arena();
+    leave_task_pool();
 }
 
 //TODO: move to arena_slot
@@ -598,9 +617,9 @@ void generic_scheduler::commit_relocated_tasks ( size_t new_tail ) {
     __TBB_ASSERT( is_local_task_pool_quiescent(),
                   "Task pool must be locked when calling commit_relocated_tasks()" );
     __TBB_store_relaxed( my_arena_slot->head, 0 );
-    // Tail is updated last to minimize probability of a thread making arena 
+    // Tail is updated last to minimize probability of a thread making arena
     // snapshot being misguided into thinking that this task pool is empty.
-    __TBB_store_relaxed( my_arena_slot->tail, new_tail );
+    __TBB_store_release( my_arena_slot->tail, new_tail );
     release_task_pool();
 }
 
@@ -652,13 +671,17 @@ inline intptr_t generic_scheduler::effective_reference_priority () const {
     // a lower priority arena, they should use arena's priority as a reference, lest
     // be trapped in a futile spinning (because market's priority would prohibit
     // executing ANY tasks in this arena).
-    return !worker_outermost_level() || 
-            my_arena->my_num_workers_allotted < my_arena->num_workers_active()
-            ? *my_ref_top_priority : my_arena->my_top_priority;
+    return !worker_outermost_level() ||
+        (my_arena->my_num_workers_allotted < my_arena->num_workers_active()
+#if __TBB_ENQUEUE_ENFORCED_CONCURRENCY
+         && my_arena->my_concurrency_mode!=arena_base::cm_enforced_global
+#endif
+            ) ? *my_ref_top_priority : my_arena->my_top_priority;
 }
 
 inline void generic_scheduler::offload_task ( task& t, intptr_t /*priority*/ ) {
     GATHER_STATISTIC( ++my_counters.prio_tasks_offloaded );
+    __TBB_ASSERT( !is_proxy(t), "The proxy task cannot be offloaded" );
     __TBB_ASSERT( my_offloaded_task_list_tail_link && !*my_offloaded_task_list_tail_link, NULL );
 #if TBB_USE_ASSERT
     t.prefix().state = task::ready;
